/* Calcula mapa de alturas a partir de trs imagens com luzes diferentes. */
/* Last edited on 2005-08-15 17:08:31 by stolfi */ 

#define PROG_NAME "calcula_alturas"
#define PROG_HELP \
  PROG_NAME " PREF_ENTRA PREF_SAIDA [ PREF_DEBUG ]"

#define _GNU_SOURCE
#include <sistema.h>
#include <imagem_vetores.h>
#include <imagem_valores.h>

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>

/* PROTTIPOS DE FUNES */

/* 
  Nos comentrios, supe-se que as linhas so numeradas DE BAIXO PARA
  CIMA e as colunas so numeradas da esquerda para a direita.
  Portanto o pixel {[0][0]}  o canto INFERIOR esquerdo da imagem.
  */

imagem_valores_t *calcula_derivadas(imagem_vetores_t *IN, char eixo);
  /* Dada a imagem de vetores normais {IN}, devolve uma imagem de
    valores {ID}, com mesmo tamanho, tal que {ID[y][x]}  a derivada
    da altura no centro do pixel {[y][x]}, isto , no ponto
    {(x+0.5,y+0.5)}. A derivada  {dz/dx} ou {dz/dy}, conforme {eixo}
    for 'x' ou 'y', respectivamente; ela  calculada a partir da
    normal {IN[y][x]} no mesmo ponto. */

imagem_valores_t *calcula_alturas_recursiva
  ( imagem_valores_t *IDX, 
    imagem_valores_t *IDY, 
    int prof,
    char *prefDebug
  );
  /* Calcula uma imagem de alturas {IZ} dadas imagens de derivadas {IDX}
    (horizontal) e {IDY} (vertical). A imagem {IZ} ter uma coluna e uma
    linha a mais do que {IDX,IDY} (gue devem ter o mesmo tamanho).

    A imagem  calculada resolvendo-se um sistema linear (vide abaixo)
    pelo mtodo de Gauss-Jordan. Este procedimento usa a tcnica
    (recursiva) de multiplas escalas para acelerar a convergncia do
    mesmo. O parmetro {prof} indica a profundidade da recurso. Se
    {prefDebug} no  NULL, escreve as imagens de derivadas e alturas
    usadas em cada escala, bem como o sistema linear correspondente. */

imagem_valores_t *reduz_derivadas(imagem_valores_t *ID);
  /* Dada uma imagem de derivadas {ID} de uma imagem de alturas {IZ},
    retorna outra imagem {JD} com as derivadas para outra imagem {JZ}
    de alturas, com metade do tamanho e com as alturas divididas pela metade. 
    
    Se o tamanho de {ID} for mpar, a ltima linha e/ou a ltima coluna 
    so duplicadas implicitamente antes da reduo. */

void escreve_sistema(sistema_t *S, char *nome, int tx, int ty);
  /* Escreve o sistema {S} no arquivo com o {nome} dado,
    explicitando os ndices dos pixels a que se refere cada equao 
    e cada incgnita. */

void preenche_sistema(imagem_valores_t *IDX, imagem_valores_t *IDY, sistema_t* S);
  /*
    Dadas duas matrizes {IDX} e {IDY} com as derivadas experimentais de
    uma imagem de alturas, preenche os coeficientes do sistema linear
    {S} cuja soluo so as tais alturas.
    
    As imagens {IDX} e {IDY} devem ter o mesmo tamanho {tx} colunas e
    {ty} linhas. O valor de {IDX[y][x]} deve ser a derivada horizontal
    {Dz/Dx} da altura, medida no centro do pixel da linha {y} e coluna
    {x}. O valor de {IDY[y][x]} deve ser a derivada vertical {Dz/Dy},
    no mesmo ponto.
    
    O sistema tem a forma {M Z = B} onde {M}  uma matriz conhecida,
    {Z}  um vetor de incgnitas, e {B} um vetor conhecido. As
    incgnitas {Z[j]} so as alturas nos vrtices (cantos dos pixels)
    de {IDX} e {IDY}; e portanto o tamanho do sistema  {N =
    (tx+1)+(ty+1)}.
    
    Mais precisamente, a incgnita {Z[j]}, para todo {j = 0..N-1},  a
    altura no ponto {(x,y)} tal que {j = x + (tx+1)*y}, com {x} em
    {0..tx} e {y} em {0..ty}. Este ponto  o canto inferior esquerdo
    do pixel {IDX[y][x]}. */
      
void extrai_valores_da_imagem(imagem_valores_t *IZ, double *Z);      
void coloca_valores_na_imagem(double *Z, imagem_valores_t *IZ);
  /* Copia os valores das incgnitas {Z[i]} de ou para os 
    elementos da imagem de alturas {IZ->elem[y][x]}. */

imagem_valores_t *expande_alturas(imagem_valores_t *JZ, int tx, int ty);
  /* Dada uma imagem de alturas {JZ}, retorna outra imagem {IZ} com
    {tx} colunas e {ty} linhas.
    
    A imagem original deve ter tamanho {tx/2 + 1} por {ty/2 + 1},
    arredondado para baixo. Na nova imagem, pixels de ndices pares
    so copiados do original, e pixels mpares so interpolados. */

int main(int argc,char** argv);

/* IMPLEMENTAES */

int main(int argc,char** argv)
  {
    /* A entrada do programa  uma imagem {IN} onde cada elemento
      {IN.pixel[y][x]}  um vetor normal associado ao centro do pixel
      na linha {y} e coluna {x} da imagem. Supe-se que o pixel  um
      quadrado de lado 1. Esses vetores so obtidos, por exemplo, pela
      comparao de trs ou mais imagens tiradas com iluminaes
      diferentes.
      
      A sada do programa  outra imagem {IZ} onde cada elemento
      {IZ.pixel[y][x]}  a altura estimada no canto inferior esquerdo
      do pixel na linha {y} e coluna {x} da imagem. Esta imagem tem
      uma linha e uma coluna a mais que a imagem de normais {IN}.
      
      A imagem {IZ}  calculada resolvendo-se um sistema linear onde o
      as incgnitas so as alturas {IZ.pixel[y][x]}.
      
      Se houver um terceiro argumento na linha de comando, o programa
      escreve tambm o sistema linear usado para encontrar as alturas,
      para depurao. */

    if ((argc < 3) || (argc > 4))
      { fprintf(stderr, "Chamada Correta: %s\n", PROG_HELP); return 1; }

    /* Pega prefixos para os nomes de arquivos. O nome do 
      arquivo de entrada  "{prefEntra}-normais.ivet".
      O nome do arquivo de sada  "{prefSaida}-alturas.ival".
      Os arquivos de depurao so "{prefDebug}-*.*". */
    char *prefEntra = argv[1];
    char *prefSaida = argv[2];
    char *prefDebug = (argc == 4 ? argv[3] : NULL);

    char* nomeArq; /* Temporrio para montar nomes de arquivos. */
    
    nomeArq = NULL; asprintf(&nomeArq, "%s-normais.ivet", prefEntra);
    fprintf(stderr, "Lendo a imagem de normais %s...\n", nomeArq);
    imagem_vetores_t *IN = le_imagem_vetores(nomeArq);
    free(nomeArq);
    
    fprintf(stderr, "Criando as imagens de derivadas...\n");
    imagem_valores_t* IDX = calcula_derivadas(IN, 'x');
    imagem_valores_t* IDY = calcula_derivadas(IN, 'y');

    fprintf(stderr, "Calculando a imagem de alturas...\n");
    imagem_valores_t* IZ = calcula_alturas_recursiva(IDX, IDY, 0, prefDebug);

    nomeArq = NULL; asprintf(&nomeArq, "%s-alturas.ival", prefSaida);
    fprintf(stderr, "Escrevendo a imagem de alturas %s...\n", nomeArq);
    escreve_imagem_valores(nomeArq, IZ);

    fprintf(stderr, "Feito!\n");
    return 0;
  }

#define TAM_CORTE 10
  /* No vale a pena recurso abaixo deste tamanho (acho). */

imagem_valores_t *calcula_alturas_recursiva
  ( imagem_valores_t *IDX, 
    imagem_valores_t *IDY, 
    int prof,
    char *prefDebug
  )
  {
    /* Pega tamanho das imagens e verifica compatibilidade: */
    int tx = IDX->tx; assert(tx == IDY->tx);
    int ty = IDX->ty; assert(ty == IDY->ty);

    int ind = 2*prof+2; /* Indentao dos comentrios. */

    fprintf(stderr, "%*sIncio do nvel %d (IDX,IDY = %d%d)...\n", ind, "", prof, tx, ty);
    char *nomeArq;

    /* Escreve matrizes de derivadas, se foi pedido: */
    if (prefDebug != NULL)
      {
        nomeArq = NULL; asprintf(&nomeArq, "%s-%02d-dZdX.ival", prefDebug, prof);
        fprintf(stderr, "%*sEscrevendo %s...\n", ind, "", nomeArq);
        escreve_imagem_valores(nomeArq, IDX);
        free(nomeArq);
        
        nomeArq = NULL; asprintf(&nomeArq, "%s-%02d-dZdY.ival", prefDebug, prof);
        fprintf(stderr, "%*sEscrevendo %s...\n", ind, "", nomeArq);
        escreve_imagem_valores(nomeArq, IDY);
        free(nomeArq);
      }
    
    /* Cria a imagem de alturas: */
    imagem_valores_t* IZ;
    
    /* Decide se deve usar o mtodo multi-escala: */
    int pequena = ((tx < TAM_CORTE) && (ty < TAM_CORTE));
    int atomica = ((tx == 1) || (ty == 1));
    if (pequena || atomica)
      { /* Inicializa {IZ} com zeros: */
        IZ = cria_imagem_valores(tx+1, ty+1);
        int x, y;
        for(y = 0; y < IZ->ty; y++) 
          for(x = 0; x < IZ->tx; x++) 
            { IZ->pixel[y][x] = 0; }
      }
    else
      { /* Inicializa {IZ} com por multi-escala. */
      
        /* Reduz imagens de derivadas pela metade: */
        fprintf(stderr, "%*sReduzindo imagens de derivadas...\n", ind, "");
        imagem_valores_t *JDX = reduz_derivadas(IDX);
        imagem_valores_t *JDY = reduz_derivadas(IDY);
        
        /* Calcula imagem de alturas reduzida: */
        imagem_valores_t *JZ = calcula_alturas_recursiva(JDX, JDY, prof+1, prefDebug);
        
        /* Expande imagem de alturas para tamanho original: */
        fprintf(stderr, "%*sExpandindo alturas para %d%d...\n", ind, "", tx+1, ty+1);
        IZ = expande_alturas(JZ, tx+1, ty+1);
      }

    /* Monta sistema linear: */
    long int N = (tx+1)*(ty+1);
    fprintf(stderr, "%*sCriando o sistema %ld%ld...\n", ind, "", N, N);
    sistema_t* S = cria_sistema(N);
    preenche_sistema(IDX, IDY, S);

    if (prefDebug != NULL)
      {
        nomeArq = NULL; asprintf(&nomeArq, "%s-%02d.sist", prefDebug, prof);
        fprintf(stderr, "%*sEscrevendo %s...\n", ind, "", nomeArq);
        escreve_sistema(S, nomeArq, IZ->tx, IZ->ty);
        free(nomeArq);
      }

    /* Resolve o sistema, obtendo as alturas: */
    fprintf(stderr, "%*sResolvendo Sistema...\n", ind, "");
    long int max_iter = 1000;
    double tol = 0.00001;
    double *Z = (double*)malloc(sizeof(double)*S->N);
    int para = 0; /* 1 significa soluo paralela, 0 sequencial. */
    int szero = 1; /* 1 significa ajustar soma em zero, 0 deixar livre. */
    extrai_valores_da_imagem(IZ, Z);
    resolve_sistema(S, Z, max_iter, tol, para, szero);
    coloca_valores_na_imagem(Z, IZ);
    
    /* Escreve matriz parcial de alturas, se foi pedido: */
    if (prefDebug != NULL)
      {
        nomeArq = NULL; asprintf(&nomeArq, "%s-%02d-alturas.ival", prefDebug, prof);
        fprintf(stderr, "%*sEscrevendo %s...\n", ind, "", nomeArq);
        escreve_imagem_valores(nomeArq, IZ);
        free(nomeArq);
      }

    fprintf(stderr, "%*sFim do nvel %d (IZ = %d%d)...\n", ind, "", prof, IZ->tx, IZ->ty);

    return IZ;
  }

void escreve_sistema(sistema_t *S, char *nome, int tx, int ty)
  {
    FILE* arq = fopen(nome, "wt");
    assert(arq != NULL); 

    long int i;
    for(i = 0; i < S->N; i++)
      {
        /* Pega a equao {i}: */
        equacao_t *eqi = &(S->eq[i]);
        
        /* Calcula indices do pixel correspondente  incgnita {Z[i]}: */
        int xi = i%tx; int yi = i/tx;
         
        /* Imprime ndices do pixel principal: */
        fprintf(arq, "eq[%d][%d]:", xi, yi);

        /* Imprime os termos lineares da equao: */
        int k;
        for(k = 0; k < MAXCOEFS; k++)
          { /* Pega mais uma incgnita {Z[j]} que entra na equao {i}: */
            long int j = eqi->ind[k];
            /* Calcula ndices do pixel correspondente: */
            int xj = j%tx; int yj = j/tx;
            double coef_j = eqi->coef[k];
            if (k > 0) { fprintf(arq, " + "); }
            fprintf(arq, "%f*Z[%d][%d]", coef_j, xj, yj);
          }
        /* Imprime o termo independente da equao: */
        fprintf(arq, " = %f", S->eq[i].indep);
        fprintf(arq, "\n");
      }
    fclose(arq);
  }

void extrai_valores_da_imagem(imagem_valores_t *IZ, double *Z)
  {
    long int N = IZ->tx * IZ->ty;
    long int i;
    for(i = 0; i < N; i++)
      { int x = i%IZ->tx;
        int y = i/IZ->tx;
        Z[i] = IZ->pixel[y][x];
      }
  }

void coloca_valores_na_imagem(double *Z, imagem_valores_t *IZ)
  {
    long int N = IZ->tx * IZ->ty;
    long int i;
    for(i = 0; i < N; i++)
      { int x = i%IZ->tx;
        int y = i/IZ->tx;
        IZ->pixel[y][x] = Z[i];
      }
  }

imagem_valores_t *calcula_derivadas(imagem_vetores_t *IN, char eixo)
  {
    assert((eixo == 'x') || (eixo == 'y'));

    int tx = IN->tx; 
    int ty = IN->ty;
    imagem_valores_t *ID = cria_imagem_valores(tx, ty);
    
    int x,y;
    for(y =0; y < ID->ty; y++)
      { for (x = 0; x < ID->tx; x++)
          { vetor_t *v = &(IN->pixel[y][x]);
            float *dpix = &(ID->pixel[y][x]);
            /* A normal deve estar apontando para cima: */
            assert(v->z > 0.0); 
            /* Calcula derivada a partir da normal: */
            if(eixo == 'x')
              { (*dpix) = - v->x / v->z; }
            else if (eixo == 'y')
              { (*dpix) = - v->y / v->z; }
          }
      }
    return ID;
  }

imagem_valores_t *reduz_derivadas(imagem_valores_t *ID)
  {
    int txI = ID->tx; int txJ = (txI+1)/2;
    int tyI = ID->ty; int tyJ = (tyI+1)/2;
    imagem_valores_t *JD = cria_imagem_valores(txJ, tyJ);
    int xJ, yJ;
    for(yJ = 0; yJ < tyJ; yJ++)
      { for(xJ = 0; xJ < txJ; xJ++)
          { /* Pixel da imagem reduzida: */
            float *pxJ = &(JD->pixel[yJ][xJ]);
            /* Pixels da imagem original (preenchendo bordas): */
            int xI = 2*xJ, yI = 2*yJ;
            double d00 = ID->pixel[yI+0][xI+0];
            double d01 = ((xI+1 < txI) ? ID->pixel[yI+0][xI+1] : d00);
            double d10 = ((yI+1 < tyI) ? ID->pixel[yI+1][xI+0] : d00);
            double d11 = ((xI+1 < txI) && (yI+1 < tyI) ? ID->pixel[yI+1][xI+1] : d00);
            /* Tanto {Z} quanto {X} e {Y} encolheram, logo a derivada no muda: */
            (*pxJ) = (d00 + d01 + d10 + d11)/4;
          }
      }
    return JD;
  }

imagem_valores_t *expande_alturas(imagem_valores_t *JZ, int tx, int ty)
  {
    int txJ = JZ->tx; assert(tx/2 + 1 == txJ);
    int tyJ = JZ->ty; assert(ty/2 + 1 == tyJ);
    imagem_valores_t *IZ = cria_imagem_valores(tx, ty);
    int xI, yI;
    for(yI = 0; yI < ty; yI++)
      { for(xI = 0; xI < tx; xI++)
          { /* Pixel da imagem expandida: */
            float *pxI = &(IZ->pixel[yI][xI]);
            /* Pixels relevantes da imagem original: */
            int xJ0 = xI/2, xJ1 = (xI+1)/2;
            int yJ0 = yI/2, yJ1 = (yI+1)/2;
            assert(xJ1 < txJ); assert(yJ1 < tyJ);
            double d00 = JZ->pixel[yJ0][xJ0];
            double d01 = JZ->pixel[yJ0][xJ1];
            double d10 = JZ->pixel[yJ1][xJ0];
            double d11 = JZ->pixel[yJ1][xJ1];
            /* Alturas devem ser ampliadas tambm: */
            (*pxI) = (d00 + d01 + d10 + d11)/2;
          }
      }
    return IZ;
  }

void preenche_sistema(imagem_valores_t *IDX, imagem_valores_t *IDY, sistema_t* S)
  {
    /* 
      Cada equao e cada incgnita do sistema {S} correponde a um
      elemento da imagem de alturas {IZ}, ou seja, a um vrtice
      (canto de pixel) das matrizes {IDX} e {IDY}. 
      
      Por definio, o vrtice {IZ[y][x]} tem coordenadas {(x,y)}.
      Portanto o pixel {IDX[y][x]}  o quadrado de lado 1 cujo canto
      inferior esquerdo  o vrtice de ndices {[y][x]}, isto , cuja
      diagonal vai de {(x,y)} a {(x+1,y+1)}.
      
      Em geral, a equao de {S} associada ao vrtice {[y][x]} tem a
      forma {Lz[y][x] = Ld[y][x]}, onde {Lz,Ld} so duas estimativas
      do laplaciano {d2z/dx2 + d2z/dy2} da funo altura no vrtice
      {[y][x]}. A estimativa {Lz}  obtida da imagem de alturas {IZ}
      por diferenas finitas entre a altura {IZ[y][x]} no vrtice e as
      alturas de seus quatro vizinhos mais prximos, e portanto  uma
      funo linear dessas cinco incgnitas.
      
      A estimativa {Ld[y][x]}  obtida a partir das derivadas
      fornecidas {IDX} e {IDY}, calculadas nos centros dos quatro
      pixels {[y][x]} dessas imagens incidentes ao vrtice {[y][x]};
      isto , {IDX[y-dy][x-dx]} e {IDY[y-dy][x-dx]} onde {dy} e {dx}
      so 0 ou 1. Note-se que {Ld}  independente das incgnitas.
      
      Na verdade, esta descrio vale apenas para os vrtices {[y][x]}
      que esto cercados por quatro pixels nas imagens {IDX} e {IDY}.
      Para vrtices com ndices {[0][x]}, ao longo da borda inferior
      dessas imagens (menos os dois vrtices extremos {[0][0]} e
      {[0][tx]}, a equao  {Dz[0][x] = Dd[0][x]}, onde {Dz} e {Dd}
      so duas estimativas para a derivada vertical da altura no ponto
      {(x,0.5)}. A estimativa {Dz}  calculada por diferenas finitas
      entre as alturas nos dois vrtice mais prximos, isto ,
      {Dz[0][x] = IZ[1][x] - IZ[0][x]}. A estimativa {Dd}  calculada
      a partir de {IDX[y][x-dx]} e {IDY[y][x-dx]} onde {dx}  0 ou 1.
      As equaes ao longo das outras trs bordas so anlogas.
      
      No vrtice {[0][0]} (canto inferior esquerdo das imagens
      {IDX} e {IDY}), a equao  tambm {Dz[y][x] = Dd[y][x]}, mas as
      estimativas se referem  derivada da altura no ponto {(0.5,0.5)}
      e ao longo do vetor {(1,1)}. Ou seja {Dz[0][0] = IZ[1][1] -
      IZ[0][0]}, e {Dd[0][0] = IDX[0][0] + IDY[0][0]}. As equaes
      para os outros trs vrtices ({[0][tx]}, {[ty][0]} e {[ty][tx]})
      so anlogas.
      
      Este sistema  resolvido pelo mtodo de Gauss-Jordan. Este
      sistema  indeterminado pois existe uma soluo no nula para o
      sistema homogneo {M Z = 0}, que  todas as alturas iguais a 1.
      (Note que toda linha de {M} tem soma zero, pois  uma estimativa
      da derivada ou do laplaciano da altura.)
    */

    /* Obtem o tamanho das imagens de derivadas e verifica consistncia: */
    int tx = IDX->tx; assert(tx == IDY->tx);
    int ty = IDX->ty; assert(ty == IDY->ty);
    
    /* Verifica tamanho do sistema: */
    assert(S->N == (tx+1)*(ty+1));
    
    auto long int indZ(int y, int x); 
      /* Calcula o ndice do vrtice {[y][x]} na imagem {IZ}. */
    
    long int indZ(int y, int x) { return x + y*(tx+1); }
    
    /* Garante que o sistema  representvel: */
    assert(MAXCOEFS >= 5);

    /* Enumera vrtices das imagens {IDX,IDY}, i.e. elems da imagem de alturas: */
    int x, y;
    for(y = 0; y <= ty; y++)
      {
        for(x = 0; x <= tx; x++)
          {
            /* Calcula ndice {i} da equao associada  altura {IZ[y][x]}: */
            long int i = indZ(y, x); 
            
            /* Pega a equao {i} e limpa a mesma: */
            equacao_t *eqi = &(S->eq[i]);
            int k;
            for (k = 0; k < MAXCOEFS; k++) 
              { eqi->coef[k] = 0.0; eqi->ind[k] = -1; }
            eqi->indep = 0.0;

            /* Monta a equao: */
            if ((y == 0) && (x == 0))
              { /* Equao para vrtice de canto {[0][0]}: */
                eqi->ind[0] = i;               eqi->coef[0] = - 1.00;
                eqi->ind[1] = indZ(1,1);       eqi->coef[1] = + 1.00;
                eqi->indep = + IDX->pixel[0][0] + IDY->pixel[0][0];
              }
            else if ((y == 0) && (x == tx))
              { /* Equao para vrtice de canto {[0][tx]}: */
                eqi->ind[0] = i;                eqi->coef[0] = - 1.00;
                eqi->ind[1] = indZ(1,tx-1);     eqi->coef[1] = + 1.00;
                eqi->indep = - IDX->pixel[0][tx-1] + IDY->pixel[0][tx-1];
              }
            else if ((y == ty) && (x == 0))
              { /* Equao para vrtice de canto {[ty][0]}: */
                eqi->ind[0] = i;                eqi->coef[0] = - 1.00;
                eqi->ind[1] = indZ(ty-1,1);     eqi->coef[1] = + 1.00;
                eqi->indep = + IDX->pixel[ty-1][0] - IDY->pixel[ty-1][0];
              }
            else if ((y == ty) && (x == tx))
              { /* Equao para vrtice de canto {[ty][tx]}: */
                eqi->ind[0] = i;                eqi->coef[0] = - 1.00;
                eqi->ind[1] = indZ(ty-1,tx-1);  eqi->coef[1] = + 1.00;
                eqi->indep = - IDX->pixel[ty-1][tx-1] - IDY->pixel[ty-1][tx-1];
              }
            
            else if (y == 0)
              { /* Equao para vrtice na borda {[0][x]}: */
                eqi->ind[0] = i;                eqi->coef[0] = - 1.00;
                eqi->ind[1] = indZ(1,x);        eqi->coef[1] = + 1.00;
                eqi->indep = + (IDY->pixel[0][x-1] + IDY->pixel[0][x])/2;
              }
            else if (y == ty)
              { /* Equao para vrtice na borda {[ty][x]}: */
                eqi->ind[0] = i;                eqi->coef[0] = - 1.00;
                eqi->ind[1] = indZ(ty-1,x);     eqi->coef[1] = + 1.00;
                eqi->indep = - (IDY->pixel[ty-1][x-1] + IDY->pixel[ty-1][x])/2;
              }
            else if (x == 0)
              { /* Equao para vrtice na borda {[y][0]}: */
                eqi->ind[0] = i;                eqi->coef[0] = - 1.00;
                eqi->ind[1] = indZ(y,1);        eqi->coef[1] = + 1.00;
                eqi->indep = + (IDX->pixel[y-1][0] + IDX->pixel[y][0])/2;
              }
            else if (x == tx)
              { /* Equao para vrtice na borda {[y][tx]}: */
                eqi->ind[0] = i;                eqi->coef[0] = - 1.00;
                eqi->ind[1] = indZ(y,tx-1);     eqi->coef[1] = + 1.00;
                eqi->indep = - (IDX->pixel[y-1][tx-1] + IDX->pixel[y][tx-1])/2;
              }
            else
              { /* Equao para vrtice interior {[y][x]}: */
                eqi->ind[0] = i;                eqi->coef[0] = - 4.00;
                eqi->ind[1] = indZ(y-1,x);      eqi->coef[1] = + 1.00;
                eqi->ind[2] = indZ(y+1,x);      eqi->coef[2] = + 1.00;
                eqi->ind[3] = indZ(y,x-1);      eqi->coef[3] = + 1.00;
                eqi->ind[4] = indZ(y,x+1);      eqi->coef[4] = + 1.00;
                eqi->indep = 
                  ( IDX->pixel[y-1][x] - IDX->pixel[y-1][x-1] +
                    IDX->pixel[y][x]   - IDX->pixel[y][x-1]   +
                    IDY->pixel[y][x-1] - IDY->pixel[y-1][x-1] +
                    IDY->pixel[y][x]   - IDY->pixel[y-1][x]
                  ) / 2;
              }
          }
      }
  }

